#if HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#ifdef __powerpc__
#include <sys/ptrace.h>
#endif

#include "ltrace.h"
#include "options.h"
#include "debug.h"
#include "dict.h"
#include "elf.h"

void
breakpoint_on_hit(struct breakpoint *bp, struct process *proc)
{
	assert(bp != NULL);
	if (bp->cbs != NULL && bp->cbs->on_hit != NULL)
		(bp->cbs->on_hit) (bp, proc);
}

void
breakpoint_on_destroy(struct breakpoint *bp)
{
	assert(bp != NULL);
	if (bp->cbs != NULL && bp->cbs->on_destroy != NULL)
		(bp->cbs->on_destroy) (bp);
}

/*****************************************************************************/

struct breakpoint *address2bpstruct(struct process *proc, void *addr)
{
	assert(proc != NULL);
	assert(proc->breakpoints != NULL);
	assert(proc->leader == proc);
	if (!proc->breakpoints) {
		proc->breakpoints =
		  dict_init(dict_key2hash_int, dict_key_cmp_int);
	}
	return dict_find_entry(proc->breakpoints, addr);
}

struct breakpoint *
insert_breakpoint(struct process *proc, void *addr,
		  struct library_symbol *libsym, int enable)
{
	struct breakpoint *sbp;
	debug(1, "insert_breakpoint(symbol=%s, addr=%p)", libsym?libsym->name:"(nil)", addr);

	struct process * leader = proc->leader;

	/* Only the group leader should be getting the breakpoints and
	 * thus have ->breakpoint initialized.  */
	assert(leader != NULL);
	assert(leader->breakpoints != NULL);

	if (!proc->breakpoints) {
		proc->breakpoints =
		    dict_init(dict_key2hash_int, dict_key_cmp_int);
	}

	if (!addr)
		return NULL;

	if (libsym)
		libsym->needs_init = 0;

	sbp = dict_find_entry(leader->breakpoints, addr);
	if (!sbp) {
		sbp = calloc(1, sizeof(struct breakpoint));
		if (!sbp) {
			return NULL;	/* TODO FIXME XXX: error_mem */
		}
		dict_enter(leader->breakpoints, addr, sbp);
		sbp->addr = addr;
		sbp->libsym = libsym;
		if (libsym)
			libsym->brkpnt = sbp;
	}
	sbp->enabled++;
	if (sbp->enabled == 1 && enable) {
		enable_breakpoint(proc, sbp);
	}

	return sbp;
}

void delete_breakpoint(struct process *proc, void *addr)
{
	struct process * leader = proc->leader;
	assert(leader != NULL);

	struct breakpoint *sbp = dict_find_entry(leader->breakpoints, addr);
	assert(sbp);		/* FIXME: remove after debugging has been done. */
	/* This should only happen on out-of-memory conditions. */
	if (sbp == NULL)
		return;

	sbp->enabled--;
	if (sbp->enabled == 0)
		disable_breakpoint(proc, sbp);
	assert(sbp->enabled >= 0);
}

static void enable_bp_cb(void *addr, void *sbp, void *proc)
{
	if (((struct breakpoint *)sbp)->enabled) {
		enable_breakpoint(proc, sbp);
	}
}

void enable_all_breakpoints(struct process *proc)
{
	debug(1, "Enabling breakpoints for pid %u...", proc->pid);
	if (proc->breakpoints) {
		dict_apply_to_all(proc->breakpoints, enable_bp_cb,
				  proc);
	}
}

static void disable_bp_cb(void *addr, void *sbp, void *proc)
{
	if (((struct breakpoint *)sbp)->enabled) {
		disable_breakpoint(proc, sbp);
	}
}

void disable_all_breakpoints(struct process *proc)
{
	assert(proc->leader == proc);
	debug(1, "Disabling breakpoints for pid %u...", proc->pid);
	dict_apply_to_all(proc->breakpoints, disable_bp_cb, proc);
}

static void free_bp_cb(void *addr, void *sbp, void *data)
{
	assert(sbp);
	free(sbp);
}

static void
entry_callback_hit(struct breakpoint *bp, struct process *proc)
{
	if (proc == NULL || proc->leader == NULL)
		return;
	delete_breakpoint(proc, bp->addr); // xxx
	reinitialize_breakpoints(proc->leader);
}

int
breakpoints_init(struct process *proc, int enable)
{
	if (proc->breakpoints) {	/* let's remove that struct */
		/* TODO FIXME XXX: free() all "struct breakpoint"s */
		dict_apply_to_all(proc->breakpoints, free_bp_cb, NULL);
		dict_clear(proc->breakpoints);
		proc->breakpoints = NULL;
	}

	/* Only the thread group leader should hold the breakpoints.
	 * (N.B. PID may be set to 0 temporarily when called by
	 * handle_exec).  */
	assert(proc->leader == proc);

	proc->breakpoints = dict_init(dict_key2hash_int,
				      dict_key_cmp_int);

	destroy_library_symbol_chain(proc->list_of_symbols);
	proc->list_of_symbols = NULL;

	GElf_Addr entry;
	if (opt_L && proc->filename) {
		proc->list_of_symbols = read_elf(proc, &entry);
		if (proc->list_of_symbols == NULL) {
		fail:
			/* XXX leak breakpoints */
			return -1;
		}

		if (opt_e) {
			struct library_symbol **tmp1 = &proc->list_of_symbols;
			while (*tmp1) {
				struct opt_e_t *tmp2 = opt_e;
				int keep = !opt_e_enable;

				while (tmp2) {
					if (!strcmp((*tmp1)->name,
						    tmp2->name)) {
						keep = opt_e_enable;
					}
					tmp2 = tmp2->next;
				}
				if (!keep) {
					*tmp1 = (*tmp1)->next;
				} else {
					tmp1 = &((*tmp1)->next);
				}
			}
		}
	}
	struct breakpoint *entry_bp
		= insert_breakpoint(proc, (void *)(uintptr_t)entry, NULL, 1);
	if (entry_bp == NULL) {
		fprintf(stderr, "fail!\n");
		goto fail;
	}

	static struct bp_callbacks entry_callbacks = {
		.on_hit = entry_callback_hit,
	};

	entry_bp->cbs = &entry_callbacks;
	proc->callstack_depth = 0;
	return 0;
}

void reinitialize_breakpoints(struct process *proc)
{
	struct library_symbol *sym = proc->list_of_symbols;

	while (sym) {
		if (sym->needs_init) {
			insert_breakpoint(proc, sym2addr(proc, sym), sym, 1);
			if (sym->needs_init && !sym->is_weak) {
				fprintf(stderr,
					"could not re-initialize breakpoint for \"%s\" in file \"%s\"\n",
					sym->name, proc->filename);
				exit(1);
			}
		}
		sym = sym->next;
	}
}
