#ifndef _HCK_LTRACE_H
#define _HCK_LTRACE_H

#include <sys/types.h>
#include <sys/time.h>
#include <stdio.h>
#include "config.h"

#include "defs.h"
#include "dict.h"

/* BREAKPOINT_LENGTH is defined in "sysdep.h" */
#include "sysdep.h"

#define MAX_LIBRARY	200
#define TRACE_FORK	1

#if defined HAVE_LIBIBERTY || defined HAVE_LIBSUPC__
# define USE_DEMANGLE
#endif

extern char *command;

extern int exiting;		/* =1 if we have to exit ASAP */

struct breakpoint;
struct process;

struct bp_callbacks {
	void (*on_hit) (struct breakpoint *bp, struct process *proc);
	void (*on_destroy) (struct breakpoint *bp);
};

struct breakpoint {
	struct bp_callbacks *cbs;
	void *addr;
	unsigned char orig_value[BREAKPOINT_LENGTH];
	int enabled;
	struct library_symbol *libsym;
};

/* Call on-hit handler of BP, if any is set.  */
void breakpoint_on_hit(struct breakpoint *bp, struct process *proc);

/* Call on-destroy handler of BP, if any is set.  */
void breakpoint_on_destroy(struct breakpoint *bp);

enum arg_type {
	ARGTYPE_UNKNOWN = -1,
	ARGTYPE_VOID,
	ARGTYPE_INT,
	ARGTYPE_UINT,
	ARGTYPE_LONG,
	ARGTYPE_ULONG,
	ARGTYPE_OCTAL,
	ARGTYPE_CHAR,
	ARGTYPE_ADDR,
	ARGTYPE_FILE,
	ARGTYPE_FORMAT,		/* printf-like format */
	ARGTYPE_STRING,
	ARGTYPE_STRING0,	/* stringN: string up to (arg N) bytes */
	ARGTYPE_STRING1,
	ARGTYPE_STRING2,
	ARGTYPE_STRING3,
	ARGTYPE_STRING4,
	ARGTYPE_STRING5
};

enum tof {
	LT_TOF_NONE = 0,
	LT_TOF_FUNCTION,	/* A real library function */
	LT_TOF_FUNCTIONR,	/* Return from a real library function */
	LT_TOF_SYSCALL,		/* A syscall */
	LT_TOF_SYSCALLR		/* Return from a syscall */
};

struct function {
	const char *name;
	enum arg_type return_type;
	int num_params;
	enum arg_type arg_types[MAX_ARGS];
	int params_right;
	struct function *next;
};

enum toplt {
	LS_TOPLT_NONE = 0,	/* PLT not used for this symbol. */
	LS_TOPLT_EXEC,		/* PLT for this symbol is executable. */
	LS_TOPLT_POINT		/* PLT for this symbol is a non-executable. */
};


extern struct function *list_of_functions;
extern char *PLTs_initialized_by_here;

struct library_symbol {
	char *name;
	void *enter_addr;
	struct breakpoint *brkpnt;
	char needs_init;
	enum toplt plt_type;
	char is_weak;
	char is_indirect;
	struct library_symbol *next;
};

struct callstack_element {
	union {
		int syscall;
		struct library_symbol *libfunc;
	} c_un;
	int is_syscall;
	void *return_addr;
	struct timeval time_spent;
};

#define MAX_CALLDEPTH 64

typedef enum Process_State Process_State;
enum Process_State {
	STATE_ATTACHED,
	STATE_BEING_CREATED,
	STATE_IGNORED  /* ignore this process (it's a fork and no -f was used) */
};

typedef struct Event_Handler Event_Handler;
struct Event_Handler {
	/* Event handler that overrides the default one.  Should
	 * return NULL if the event was handled, otherwise the
	 * returned event is passed to the default handler.  */
	struct event * (* on_event)(Event_Handler * self, struct event * event);

	/* Called when the event handler removal is requested.  */
	void (* destroy)(Event_Handler * self);
};

/* XXX We would rather have this all organized a little differently,
 * have Process for the whole group and Task for what's there for
 * per-thread stuff.  But for now this is the less invasive way of
 * structuring it.  */
struct process {
	Process_State state;
	struct process *parent; /* needed by STATE_BEING_CREATED */
	char *filename;
	char *interp;
	pid_t pid;

	/* Dictionary of breakpoints (which is a mapping
	 * address->struct breakpoint).  This is NULL for non-leader
	 * processes.  */
	struct dict *breakpoints;

	int mask_32bit;		/* 1 if 64-bit ltrace is tracing 32-bit process.  */
	unsigned int personality;
	int tracesysgood;	/* signal indicating a PTRACE_SYSCALL trap */
	int early;	/* for consistency checks, this is true for
			 * children whose TRAP was delivered before
			 * the fork message of the parent.  */

	int seen_initial;	/* whether we've seen the trap for
				 * first instruction after exec.  */

	int callstack_depth;
	struct callstack_element callstack[MAX_CALLDEPTH];
	struct library_symbol *list_of_symbols;

	int libdl_hooked;
	/* Arch-dependent: */
	void * debug;	/* arch-dep process debug struct */
	long debug_state; /* arch-dep debug state */
	void *instruction_pointer;
	void *stack_pointer;	/* To get return addr, args... */
	void *return_addr;
	void *dyn_addr;
	void *arch_ptr;
	short e_machine;
	short need_to_reinitialize_breakpoints;

	/* Set in leader.  */
	Event_Handler * event_handler;


	/**
	 * Process chaining.
	 **/
	struct process *next;

	/* LEADER points to the leader thread of the POSIX.1 process.
	   If X->LEADER == X, then X is the leader thread and the
	   Process structures chained by NEXT represent other threads,
	   up until, but not including, the next leader thread.
	   LEADER may be NULL after the leader has already exited.  In
	   that case this process is waiting to be collected.  */
	struct process *leader;
};

struct event {
	struct event * next;
	struct process *proc;
	enum {
		LT_EV_UNKNOWN,
		LT_EV_NONE,
		LT_EV_SIGNAL,
		LT_EV_EXIT,
		LT_EV_EXIT_SIGNAL,
		LT_EV_SYSCALL,
		LT_EV_SYSRET,
		LT_EV_CLONE,
		LT_EV_VFORK,
		LT_EV_EXEC,
		LT_EV_BREAKPOINT,
		LT_EV_NEW
	} thing;
	union {
		int ret_val;	/* _EV_EXIT */
		int signum;	/* _EV_SIGNAL, _EV_EXIT_SIGNAL */
		int sysnum;	/* _EV_SYSCALL, _EV_SYSRET */
		void *brk_addr;	/* _EV_BREAKPOINT */
		int newpid;	/* _EV_CLONE, _EV_NEW */
	} e_un;
};

struct opt_c_struct {
	int count;
	struct timeval tv;
};
extern struct dict *dict_opt_c;

enum process_status {
	ps_invalid,	/* Failure.  */
	ps_stop,	/* Job-control stop.  */
	ps_tracing_stop,
	ps_sleeping,
	ps_zombie,
	ps_other,	/* Necessary other states can be added as needed.  */
};

enum pcb_status {
	pcb_stop, /* The iteration should stop.  */
	pcb_cont, /* The iteration should continue.  */
};

extern void *instruction_pointer;

/* Events  */
enum ecb_status {
	ecb_cont, /* The iteration should continue.  */
	ecb_yield, /* The iteration should stop, yielding this
		    * event.  */
	ecb_deque, /* Like ecb_stop, but the event should be removed
		    * from the queue.  */
};
extern struct event * each_qd_event(enum ecb_status (* cb)(struct event * ev,
							   void * data),
				    void * data);
extern void enque_event(struct event * event);
extern struct event *wait_for_something(void);

extern void install_event_handler(struct process * proc, Event_Handler * handler);
extern void destroy_event_handler(struct process * proc);

extern void process_event(struct event *event);
extern pid_t execute_program(const char * command, char ** argv);
extern int display_arg(enum tof type, struct process *proc, int arg_num,
		       enum arg_type at);
extern struct breakpoint *address2bpstruct(struct process *proc, void *addr);
extern int breakpoints_init(struct process *proc, int enable);
extern struct breakpoint *insert_breakpoint(struct process *proc, void *addr,
					    struct library_symbol *libsym,
					    int enable);
extern void delete_breakpoint(struct process *proc, void *addr);
extern void enable_all_breakpoints(struct process *proc);
extern void disable_all_breakpoints(struct process *proc);
extern void reinitialize_breakpoints(struct process *);

extern struct process *open_program(char *filename, pid_t pid,
				    int init_breakpoints);
extern void open_pid(pid_t pid, int verbose);
extern void open_forked_pid(pid_t pid, int early);
extern void show_summary(void);

extern struct library_symbol * clone_library_symbol(struct library_symbol * s);
extern void destroy_library_symbol(struct library_symbol * s);
extern void destroy_library_symbol_chain(struct library_symbol * chain);

struct ltelf;
extern int do_init_elf(struct ltelf *lte, const char *filename);
extern void do_close_elf(struct ltelf *lte);
extern struct library_symbol *library_symbols;

/* Arch-dependent stuff: */
extern char *pid2name(pid_t pid);
extern pid_t process_leader(pid_t pid);
extern int process_tasks(pid_t pid, pid_t **ret_tasks, size_t *ret_n);
extern int process_stopped(pid_t pid);
extern void wait_for_proc(pid_t pid);
extern enum process_status process_status(pid_t pid);
extern void trace_set_options(struct process *proc, pid_t pid);
extern void trace_me(void);
extern int trace_pid(pid_t pid);
extern void untrace_pid(pid_t pid);
extern void get_arch_dep(struct process *proc);
extern void *get_instruction_pointer(struct process *proc);
extern void set_instruction_pointer(struct process *proc, void *addr);
extern void *get_stack_pointer(struct process *proc);
extern void *get_return_addr(struct process *proc, void *stack_pointer);
extern void enable_breakpoint(struct process *proc, struct breakpoint *sbp);
extern void disable_breakpoint(struct process *proc, const struct breakpoint *sbp);
extern int syscall_p(struct process *proc, int status, int *sysnum);
extern void continue_process(pid_t pid);
extern void continue_after_signal(pid_t pid, int signum);
extern void continue_after_syscall(struct process *proc, int sysnum, int ret_p);
extern void continue_after_breakpoint(struct process *proc, struct breakpoint *sbp);
extern void continue_after_vfork(struct process * proc);
extern void continue_enabling_breakpoint(struct process *proc,
					 struct breakpoint *sbp);
extern long gimme_arg(enum tof type, struct process *proc, int arg_num);
extern void save_register_args(enum tof type, struct process *proc);
extern int umovestr(struct process *proc, void *addr, int len, void *laddr);
extern size_t umovebytes (struct process *proc, void * addr, void * laddr,
			  size_t count);
extern int ffcheck(void *maddr);
extern void *sym2addr(struct process *, struct library_symbol *);
extern int linkmap_init(struct process *);

extern void arch_check_dbg(struct process *proc);
extern int task_kill (pid_t pid, int sig);
extern struct process *pid2proc(int pid);
extern void add_process(struct process * proc);
extern void remove_process(struct process * proc);
extern void change_process_leader(struct process * proc, struct process * leader);
extern struct process *each_process(struct process * start,
				    enum pcb_status (* cb)(struct process *,
							   void * data),
				    void * data);
extern struct process *each_task(struct process * start,
				 enum pcb_status (* cb)(struct process * proc,
							void * data),
				 void * data);

/* These three functions cooperate in installing process exit handler.
 * LTRACE_EXITING is a back end function that is called to do the
 * actual installation.  SHOULD_CALL_LTRACE_EXITING is called by a
 * signal handler in ltrace.c to determine whether it is safe (and
 * desired) to install a process exit handler right away, or whether
 * it should be postponed and done in PROCESS_EVENT.
 * CHECK_LTRACE_EXITING is a utility wrapper that calls
 * LTRACE_EXITING.  */
int should_call_ltrace_exiting(void);
void check_ltrace_exiting(void);
void ltrace_exiting(void);

extern ulong get_child_pid(pid_t pid);
#endif
