#if HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <assert.h>
#include <unistd.h>
#include <error.h>

#include "ltrace.h"
#include "options.h"
#include "elf.h"

struct process *
open_program(char *filename, pid_t pid, int enable)
{
	struct process *proc;
	assert(pid != 0);
	proc = calloc(sizeof(struct process), 1);
	if (!proc) {
		perror("malloc");
		exit(1);
	}

	proc->filename = strdup(filename);
	proc->pid = pid;

	add_process(proc);
	if (proc->leader == NULL) {
		free(proc);
		return NULL;
	}

	if (proc->leader == proc) {
		trace_set_options(proc, proc->pid);
		if (breakpoints_init(proc, enable)) {
			fprintf(stderr, "failed to init breakpoints %d\n",
				proc->pid);
			remove_process(proc);
			return NULL;
		}
	}

	return proc;
}

static int
open_one_pid(pid_t pid)
{
	struct process *proc;
	char *filename;

	/* Get the filename first.  Should the trace_pid fail, we can
	 * easily free it, untracing is more work.  */
	if ((filename = pid2name(pid)) == NULL
	    || trace_pid(pid) < 0) {
		free(filename);
		return -1;
	}

	proc = open_program(filename, pid, 0);
	if (proc == NULL)
		return -1;
	trace_set_options(proc, pid);

	return 0;
}

static enum pcb_status
start_one_pid(struct process * proc, void * data)
{
	continue_process(proc->pid);
	return pcb_cont;
}

void
open_pid(pid_t pid, int verbose)
{
	/* If we are already tracing this guy, we should be seeing all
	 * his children via normal tracing route.  */
	if (pid2proc(pid) != NULL)
		return;

	/* First, see if we can attach the requested PID itself.  */
	if (open_one_pid(pid)) {
		fprintf(stderr, "Cannot attach to pid %u: %s\n",
			pid, strerror(errno));
		return;
	}

	/* Now attach to all tasks that belong to that PID.  There's a
	 * race between process_tasks and open_one_pid.  So when we
	 * fail in open_one_pid below, we just do another round.
	 * Chances are that by then that PID will have gone away, and
	 * that's why we have seen the failure.  The processes that we
	 * manage to open_one_pid are stopped, so we should eventually
	 * reach a point where process_tasks doesn't give any new
	 * processes (because there's nobody left to produce
	 * them).  */
	size_t old_ntasks = 0;
	int have_all;
	while (1) {
		pid_t *tasks;
		size_t ntasks;
		size_t i;

		if (process_tasks(pid, &tasks, &ntasks) < 0) {
			fprintf(stderr, "Cannot obtain tasks of pid %u: %s\n",
				pid, strerror(errno));
			goto start;
		}

		have_all = 1;
		for (i = 0; i < ntasks; ++i)
			if (pid2proc(tasks[i]) == NULL
			    && open_one_pid(tasks[i]))
				have_all = 0;

		free(tasks);

		if (have_all && old_ntasks == ntasks)
			break;
		old_ntasks = ntasks;
	}

	/* Done.  Now initialize breakpoints and then continue
	 * everyone.  */
	struct process * leader;
start:
	leader = pid2proc(pid)->leader;
	reinitialize_breakpoints(leader);

	linkmap_init(leader);
	leader->libdl_hooked = 1;

	each_task(leader, start_one_pid, NULL);
}

void open_forked_pid(pid_t pid, int early)
{
	char *filename = pid2name(pid);
	struct process *proc = open_program(filename, pid, 1);
	proc->early = early;
#ifdef __powerpc__
	breakpoints_init(proc, 1);
#endif
}

static struct process * list_of_processes = NULL;

static void
unlist_process(struct process * proc)
{
	struct process *tmp;

	if (list_of_processes == proc) {
		list_of_processes = list_of_processes->next;
		return;
	}

	for (tmp = list_of_processes; ; tmp = tmp->next) {
		/* If the following assert fails, the process wasn't
		 * in the list.  */
		assert(tmp->next != NULL);

		if (tmp->next == proc) {
			tmp->next = tmp->next->next;
			return;
		}
	}
}

struct process *
each_process(struct process * proc,
	     enum pcb_status (* cb)(struct process * proc, void * data),
	     void * data)
{
	struct process * it = proc ?: list_of_processes;
	for (; it != NULL; ) {
		/* Callback might call remove_process.  */
		struct process * next = it->next;
		if ((*cb) (it, data) == pcb_stop)
			return it;
		it = next;
	}
	return NULL;
}

struct process *
each_task(struct process * it,
	  enum pcb_status (* cb)(struct process * proc, void * data),
	  void * data)
{
	if (it != NULL) {
		struct process * leader = it->leader;
		for (; it != NULL && it->leader == leader; ) {
			/* Callback might call remove_process.  */
			struct process * next = it->next;
			if ((*cb) (it, data) == pcb_stop)
				return it;
			it = next;
		}
	}
	return NULL;
}

void
add_process(struct process * proc)
{
	struct process ** leaderp = &list_of_processes;
	if (proc->pid) {
		pid_t tgid = process_leader(proc->pid);
		if (tgid == 0)
			/* Must have been terminated before we managed
			 * to fully attach.  */
			return;
		if (tgid == proc->pid)
			proc->leader = proc;
		else {
			struct process * leader = pid2proc(tgid);
			proc->leader = leader;
			if (leader != NULL)
				leaderp = &leader->next;
		}
	}
	proc->next = *leaderp;
	*leaderp = proc;
}

void
change_process_leader(struct process * proc, struct process * leader)
{
	struct process ** leaderp = &list_of_processes;
	if (proc->leader == leader)
		return;

	assert(leader != NULL);
	unlist_process(proc);
	if (proc != leader)
		leaderp = &leader->next;

	proc->leader = leader;
	proc->next = *leaderp;
	*leaderp = proc;
}

static enum pcb_status
clear_leader(struct process * proc, void * data)
{
	proc->leader = NULL;
	return pcb_cont;
}

static enum ecb_status
event_for_proc(struct event * event, void * data)
{
	if (event->proc == data)
		return ecb_deque;
	else
		return ecb_cont;
}

static void
delete_events_for(struct process * proc)
{
	struct event * event;
	while ((event = each_qd_event(&event_for_proc, proc)) != NULL)
		free(event);
}

void
remove_process(struct process *proc)
{
	if (proc->leader == proc)
		each_task(proc, &clear_leader, NULL);

	unlist_process(proc);
	delete_events_for(proc);
	free(proc);
}

void
install_event_handler(struct process * proc, Event_Handler * handler)
{
	assert(proc->event_handler == NULL);
	proc->event_handler = handler;
}

void
destroy_event_handler(struct process * proc)
{
	Event_Handler * handler = proc->event_handler;
	assert(handler != NULL);
	if (handler->destroy != NULL)
		handler->destroy(handler);
	free(handler);
	proc->event_handler = NULL;
}
