#if HAVE_CONFIG_H
#include "config.h"
#endif

#define _GNU_SOURCE
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <assert.h>
#include <sys/time.h>
#include <errno.h>

#include "ltrace.h"
#include "output.h"
#include "options.h"
#include "elf.h"
#include "debug.h"

#ifdef __powerpc__
#include <sys/ptrace.h>
#endif

static void process_signal(struct event *event);
static void process_exit(struct event *event);
static void process_exit_signal(struct event *event);
static void process_syscall(struct event *event);
static void process_sysret(struct event *event);
static void process_clone(struct event *event);
static void process_exec(struct event *event);
static void process_breakpoint(struct event *event);

static void callstack_push_syscall(struct process *proc, int sysnum);
static void callstack_push_symfunc(struct process *proc,
				   struct library_symbol *sym);
static void callstack_pop(struct process *proc);

/* TODO */
static int clone_breakpoints(struct process * proc, struct process * orig_proc);
void * address_clone(void * addr, void * data) {
	return addr;
}

void * breakpoint_clone(void * bp, void * data) {
	struct breakpoint * b;
	struct dict * map = data;
	b = malloc(sizeof(struct breakpoint));
	if (!b) {
		perror("malloc()");
		exit(1);
	}
	memcpy(b, bp, sizeof(struct breakpoint));
	if (b->libsym != NULL) {
		struct library_symbol * sym = dict_find_entry(map, b->libsym);
		if (b->libsym == NULL) {
			fprintf(stderr, "Can't find cloned symbol %s.\n",
				b->libsym->name);
			return NULL;
		}
		b->libsym = sym;
	}
	return b;
}

typedef struct Pending_New Pending_New;

struct Pending_New {
	pid_t pid;
	Pending_New * next;
};

static Pending_New * pending_news = NULL;

static int pending_new(pid_t pid) {
	Pending_New * p = pending_news;
	while (p) {
		if (p->pid == pid) {
			return 1;
		}
		p = p->next;
	}
	return 0;
}

static void pending_new_insert(pid_t pid) {
	Pending_New * p = malloc(sizeof(Pending_New));
	if (!p) {
		perror("malloc()");
		exit(1);
	}
	p->pid = pid;
	p->next = pending_news;
	pending_news = p;
}

static void pending_new_remove(pid_t pid) {
	Pending_New *p, *pred;
	p = pending_news;
	if (p->pid == pid) {
		pending_news = p->next;
		free(p);
	} else {
		while (p) {
			Pending_New *next = p->next;
			if (p->pid == pid) {
				pred->next = p->next;
				free(p);
			}
			pred = p;
			p = next;
		}
	}
}

static void process_clone(struct event * event) {
	struct process *p;

	p = malloc(sizeof(struct process));
	if (!p) {
		perror("malloc()");
		exit(1);
	}
	memcpy(p, event->proc, sizeof(struct process));
	p->pid = event->e_un.newpid;
	p->parent = event->proc;

	/* We save register values to the arch pointer, and these need
	   to be per-thread.  */
	p->arch_ptr = NULL;

	if (pending_new(p->pid)) {
		pending_new_remove(p->pid);
		if (p->event_handler != NULL)
			destroy_event_handler(p);
		if (event->proc->state == STATE_ATTACHED && opt_f) {
			p->state = STATE_ATTACHED;
		} else {
			p->state = STATE_IGNORED;
		}
		continue_process(p->pid);
		add_process(p);
	} else {
		p->state = STATE_BEING_CREATED;
		add_process(p);
	}

	if (p->leader == p)
		clone_breakpoints(p, event->proc->leader);
	else
		/* Thread groups share breakpoints.  */
		p->breakpoints = NULL;

	if (event->thing == LT_EV_VFORK)
		continue_after_vfork(p);
	else
		continue_process(event->proc->pid);
}

static void process_new(struct event * event) {
	struct process * proc = pid2proc(event->e_un.newpid);
	if (!proc) {
		pending_new_insert(event->e_un.newpid);
	} else {
		assert(proc->state == STATE_BEING_CREATED);
		if (opt_f) {
			proc->state = STATE_ATTACHED;
		} else {
			proc->state = STATE_IGNORED;
		}
		continue_process(proc->pid);
	}
}

static char *shortsignal(struct process *proc, int signum)
{
	static char *signalent0[] = {
#include "signalent.h"
	};
	static char *signalent1[] = {
#include "signalent1.h"
	};
	static char **signalents[] = { signalent0, signalent1 };
	int nsignals[] = { sizeof signalent0 / sizeof signalent0[0],
		sizeof signalent1 / sizeof signalent1[0]
	};

	if (proc->personality > sizeof signalents / sizeof signalents[0])
		abort();
	if (signum < 0 || signum >= nsignals[proc->personality]) {
		return "UNKNOWN_SIGNAL";
	} else {
		return signalents[proc->personality][signum];
	}
}

static char *sysname(struct process *proc, int sysnum)
{
	static char result[128];
	static char *syscalent0[] = {
#include "syscallent.h"
	};
	static char *syscalent1[] = {
#include "syscallent1.h"
	};
	static char **syscalents[] = { syscalent0, syscalent1 };
	int nsyscals[] = { sizeof syscalent0 / sizeof syscalent0[0],
		sizeof syscalent1 / sizeof syscalent1[0]
	};

	if (proc->personality > sizeof syscalents / sizeof syscalents[0])
		abort();
	if (sysnum < 0 || sysnum >= nsyscals[proc->personality]) {
		sprintf(result, "SYS_%d", sysnum);
		return result;
	} else {
		sprintf(result, "SYS_%s",
			syscalents[proc->personality][sysnum]);
		return result;
	}
}

static struct event *
call_handler(struct process * proc, struct event * event)
{
	assert(proc != NULL);

	Event_Handler * handler = proc->event_handler;
	if (handler == NULL)
		return event;

	return (*handler->on_event) (handler, event);
}

void
check_ltrace_exiting(void)
{
	if (exiting == 1) {
		exiting = 2;
		debug(1, "ltrace about to exit");
		ltrace_exiting();
	}
}

void process_event(struct event *event)
{
	check_ltrace_exiting();

	/* If the thread group or an individual task define an
	   overriding event handler, give them a chance to kick in.
	   We will end up calling both handlers, if the first one
	   doesn't sink the event.  */
	if (event->proc != NULL) {
		event = call_handler(event->proc, event);
		if (event == NULL)
			/* It was handled.  */
			return;

		/* Note: the previous handler has a chance to alter
		 * the event.  */
		if (event->proc != NULL
		    && event->proc->leader != NULL
		    && event->proc != event->proc->leader) {
			event = call_handler(event->proc->leader, event);
			if (event == NULL)
				/* It was handled.  */
				return;
		}
	}

	switch (event->thing) {
	case LT_EV_NONE:
		debug(1, "event: none");
		return;
	case LT_EV_SIGNAL:
		debug(1, "event: signal (%s [%d])",
		      shortsignal(event->proc, event->e_un.signum),
		      event->e_un.signum);
		process_signal(event);
		return;
	case LT_EV_EXIT:
		debug(1, "event: exit (%d)", event->e_un.ret_val);
		process_exit(event);
		return;
	case LT_EV_EXIT_SIGNAL:
		debug(1, "event: exit signal (%s [%d])",
		      shortsignal(event->proc, event->e_un.signum),
		      event->e_un.signum);
		process_exit_signal(event);
		return;
	case LT_EV_SYSCALL:
		debug(1, "event: syscall (%s [%d])",
		      sysname(event->proc, event->e_un.sysnum),
		      event->e_un.sysnum);
		process_syscall(event);
		return;
	case LT_EV_SYSRET:
		debug(1, "event: sysret (%s [%d])",
		      sysname(event->proc, event->e_un.sysnum),
		      event->e_un.sysnum);
		process_sysret(event);
		return;
	case LT_EV_CLONE:
	case LT_EV_VFORK:
		debug(1, "event: clone (%u)", event->e_un.newpid);
		process_clone(event);
		return;
	case LT_EV_EXEC:
		debug(1, "event: exec()");
		process_exec(event);
		return;
	case LT_EV_BREAKPOINT:
		debug(1, "event: breakpoint");
		process_breakpoint(event);
		return;
	case LT_EV_NEW:
		debug(1, "event: new process");
		process_new(event);
		return;
	default:
		fprintf(stderr, "Error! unknown event?\n");
		exit(1);
	}
}

static void process_signal(struct event *event)
{
	if (event->proc->state != STATE_IGNORED) {
		output_line(event->proc, "--- %s (%s) ---",
			    shortsignal(event->proc, event->e_un.signum),
			    strsignal(event->e_un.signum));
	}
	continue_after_signal(event->proc->pid, event->e_un.signum);
}

static void process_exit(struct event *event)
{
	if (event->proc->state != STATE_IGNORED) {
		output_line(event->proc, "+++ exited (status %d) +++",
			    event->e_un.ret_val);
	}
	remove_process(event->proc);
}

static void process_exit_signal(struct event *event)
{
	if (event->proc->state != STATE_IGNORED) {
		output_line(event->proc, "+++ killed by %s +++",
			    shortsignal(event->proc, event->e_un.signum));
	}
	remove_process(event->proc);
}

static void process_syscall(struct event *event)
{
	if (event->proc->state != STATE_IGNORED) {
		if (opt_S) {
			output_left(LT_TOF_SYSCALL, event->proc,
				    sysname(event->proc, event->e_un.sysnum));
		}
			callstack_push_syscall(event->proc, event->e_un.sysnum);
	}
	continue_after_syscall(event->proc, event->e_un.sysnum, 0);
}

static void process_exec(struct event * event) {
	struct process * proc = event->proc;

	if (proc->state == STATE_IGNORED) {
		untrace_pid(proc->pid);
		remove_process(proc);
		return;
	}
	output_line(proc, "--- Called exec() ---");
	proc->mask_32bit = 0;
	proc->personality = 0;
	proc->arch_ptr = NULL;
	proc->seen_initial = 0;
	free(proc->filename);
	proc->filename = pid2name(proc->pid);
	breakpoints_init(proc, 0);
	proc->callstack_depth = 0;
	continue_process(proc->pid);
}

struct timeval current_time_spent;

static void calc_time_spent(struct process *proc)
{
	struct timeval tv;
	struct timezone tz;
	struct timeval diff;
	struct callstack_element *elem;

	elem = &proc->callstack[proc->callstack_depth - 1];

	gettimeofday(&tv, &tz);

	diff.tv_sec = tv.tv_sec - elem->time_spent.tv_sec;
	if (tv.tv_usec >= elem->time_spent.tv_usec) {
		diff.tv_usec = tv.tv_usec - elem->time_spent.tv_usec;
	} else {
		diff.tv_sec++;
		diff.tv_usec = 1000000 + tv.tv_usec - elem->time_spent.tv_usec;
	}
	current_time_spent = diff;
}

static void process_sysret(struct event *event)
{
	if (event->proc->state != STATE_IGNORED) {
		if (opt_T || opt_c) {
			calc_time_spent(event->proc);
		}
		assert(event->proc->callstack_depth > 0);
		unsigned d = event->proc->callstack_depth - 1;
		assert(event->proc->callstack[d].is_syscall);
		callstack_pop(event->proc);
		if (opt_S) {
			output_right(LT_TOF_SYSCALLR, event->proc,
				sysname(event->proc, event->e_un.sysnum));
		}
	}
	continue_after_syscall(event->proc, event->e_un.sysnum, 1);
}

static void process_breakpoint(struct event *event)
{
	int i, j;
	struct breakpoint *sbp;
	struct process *leader = event->proc->leader;

	/* The leader has terminated.  */
	if (leader == NULL) {
		continue_process(event->proc->pid);
		return;
	}

	debug(2, "event: breakpoint (%p)", event->e_un.brk_addr);

	void *brk_addr = event->e_un.brk_addr;
	for (i = event->proc->callstack_depth - 1; i >= 0; i--) {
		if (brk_addr == event->proc->callstack[i].return_addr) {

			struct library_symbol *libsym =
			    event->proc->callstack[i].c_un.libfunc;
#ifdef __powerpc__
			/*
			 * PPC HACK! (XXX FIXME TODO)
			 * The PLT gets modified during the first call,
			 * so be sure to re-enable the breakpoint.
			 */
			void *addr = sym2addr(event->proc, libsym);
			unsigned long a;

			if (libsym->plt_type != LS_TOPLT_POINT) {
				unsigned char break_insn[] = BREAKPOINT_VALUE;

				sbp = address2bpstruct(leader, addr);
				assert(sbp);
				a = ptrace(PTRACE_PEEKTEXT, event->proc->pid,
					   addr);

				if (memcmp(&a, break_insn, BREAKPOINT_LENGTH)) {
					sbp->enabled--;
					insert_breakpoint(event->proc, addr,
							  libsym, 1);
				}
			} else {
				sbp = libsym->brkpnt;
				assert(sbp);
				if (addr != sbp->addr) {
					insert_breakpoint(event->proc, addr,
							  libsym, 1);
				}
			}
#endif
			for (j = event->proc->callstack_depth - 1; j > i; j--) {
				callstack_pop(event->proc);
			}
			if (event->proc->state != STATE_IGNORED) {
				if (opt_T || opt_c) {
					calc_time_spent(event->proc);
				}
			}

			callstack_pop(event->proc);

			/* Pop also any other entries that seem like
			 * they are linked to the current one: they
			 * have the same return address, but were made
			 * for different symbols.  This should only
			 * happen for entry point tracing, i.e. for -x
			 * everywhere, or -x and -e on PPC64.  */
			while (event->proc->callstack_depth > 0) {
				struct callstack_element *prev;
				size_t d = event->proc->callstack_depth;
				prev = &event->proc->callstack[d - 1];
				if (prev->c_un.libfunc != libsym
				    && prev->return_addr == brk_addr) {
					callstack_pop(event->proc);
					i--;
				} else {
					break;
				}
			}

			event->proc->return_addr = event->e_un.brk_addr;
			if (event->proc->state != STATE_IGNORED) {
				output_right(LT_TOF_FUNCTIONR, event->proc,
					     event->proc->callstack[i].c_un.libfunc->name);
			}

			if (libsym->is_indirect) {
				libsym->is_indirect = 0;
				libsym->enter_addr
					= (void *)gimme_arg(LT_TOF_FUNCTIONR,
							    event->proc, -1);
				libsym->name[strlen(libsym->name)
					     - (sizeof(".IFUNC") - 1)] = 0;
				debug(1, "Resolved indirect call to %s to %p",
				      libsym->name, libsym->enter_addr);
				libsym->plt_type = LS_TOPLT_NONE;
				insert_breakpoint(event->proc,
						  libsym->enter_addr,
						  libsym, 1);
			}

			sbp = address2bpstruct(leader, event->e_un.brk_addr);
			continue_after_breakpoint(event->proc, sbp);
			return;
		}
	}

	if ((sbp = address2bpstruct(leader, event->e_un.brk_addr))) {
		breakpoint_on_hit(sbp, event->proc);
		if (sbp->libsym == NULL) {
			continue_after_breakpoint(event->proc, sbp);
			return;
		}

		if (strcmp(sbp->libsym->name, "") == 0) {
			debug(2, "Hit _dl_debug_state breakpoint!\n");
			arch_check_dbg(event->proc);
		}
		if (event->proc->state != STATE_IGNORED) {
			event->proc->stack_pointer = get_stack_pointer(event->proc);
			event->proc->return_addr =
				get_return_addr(event->proc, event->proc->stack_pointer);
			output_left(LT_TOF_FUNCTION, event->proc, sbp->libsym->name);
			callstack_push_symfunc(event->proc, sbp->libsym);
		}

		continue_after_breakpoint(event->proc, sbp);
		return;
	}

	if (event->proc->state != STATE_IGNORED) {
		if (!event->proc->seen_initial)
			event->proc->seen_initial = 1;
		else
			output_line(event->proc, "unexpected breakpoint at %p",
				    (void *)event->e_un.brk_addr);
	}
	continue_process(event->proc->pid);
}

static void callstack_push_syscall(struct process *proc, int sysnum)
{
	struct callstack_element *elem;

	/* FIXME: not good -- should use dynamic allocation. 19990703 mortene. */
	if (proc->callstack_depth == MAX_CALLDEPTH - 1) {
		fprintf(stderr, "Error: call nesting too deep!\n");
		return;
	}

	elem = &proc->callstack[proc->callstack_depth];
	elem->is_syscall = 1;
	elem->c_un.syscall = sysnum;
	elem->return_addr = NULL;

	proc->callstack_depth++;
	if (opt_T || opt_c) {
		struct timezone tz;
		gettimeofday(&elem->time_spent, &tz);
	}
}

static int
clone_breakpoints(struct process * proc, struct process * orig_proc)
{
	/* When copying breakpoints, we also have to copy the
	 * referenced symbols, and link them properly.  */
	struct dict * map = dict_init(&dict_key2hash_int, &dict_key_cmp_int);
	struct library_symbol * it = proc->list_of_symbols;
	proc->list_of_symbols = NULL;
	for (; it != NULL; it = it->next) {
		struct library_symbol * libsym = clone_library_symbol(it);
		if (libsym == NULL) {
			int save_errno;
		err:
			save_errno = errno;
			destroy_library_symbol_chain(proc->list_of_symbols);
			dict_clear(map);
			errno = save_errno;
			return -1;
		}
		libsym->next = proc->list_of_symbols;
		proc->list_of_symbols = libsym;
		if (dict_enter(map, it, libsym) != 0)
			goto err;
	}

	proc->breakpoints = dict_clone2(orig_proc->breakpoints,
					address_clone, breakpoint_clone, map);
	if (proc->breakpoints == NULL)
		goto err;

	dict_clear(map);
	return 0;
}

static void
callstack_push_symfunc(struct process *proc, struct library_symbol *sym)
{
	struct callstack_element *elem;

	/* FIXME: not good -- should use dynamic allocation. 19990703 mortene. */
	if (proc->callstack_depth == MAX_CALLDEPTH - 1) {
		fprintf(stderr, "Error: call nesting too deep!\n");
		return;
	}

	elem = &proc->callstack[proc->callstack_depth];
	elem->is_syscall = 0;
	elem->c_un.libfunc = sym;

	elem->return_addr = proc->return_addr;
        if (elem->return_addr) {
		insert_breakpoint(proc, elem->return_addr, NULL, 1);
	}

	proc->callstack_depth++;
	if (opt_T || opt_c) {
		struct timezone tz;
		gettimeofday(&elem->time_spent, &tz);
	}
}

static void callstack_pop(struct process *proc)
{
	struct callstack_element *elem;
	assert(proc->callstack_depth > 0);

	elem = &proc->callstack[proc->callstack_depth - 1];
	if (!elem->is_syscall && elem->return_addr) {
		assert(proc->leader != NULL);
		delete_breakpoint(proc, elem->return_addr);
	}
	proc->callstack_depth--;
}
