#define _GNU_SOURCE /* For getline.  */
#if HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <link.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <dirent.h>
#include <ctype.h>
#include <errno.h>
#include <sys/syscall.h>
#include "ltrace.h"
#include "debug.h"
#include "elf.h"
#include "options.h"
#include <stdlib.h>
#include <error.h>

/* /proc/pid doesn't exist just after the fork, and sometimes `ltrace'
 * couldn't open it to find the executable.  So it may be necessary to
 * have a bit delay
 */

#define	MAX_DELAY	100000	/* 100000 microseconds = 0.1 seconds */

#define PROC_PID_FILE(VAR, FORMAT, PID)		\
	char VAR[strlen(FORMAT) + 6];		\
	sprintf(VAR, FORMAT, PID)

/*
 * Returns a file name corresponding to a running pid
 */
char *pid2name(pid_t pid)
{
	if (!kill(pid, 0)) {
		int delay = 0;

		PROC_PID_FILE(proc_exe, "/proc/%d/exe", pid);

		while (delay < MAX_DELAY) {
			if (!access(proc_exe, F_OK)) {
				return strdup(proc_exe);
			}
			delay += 1000;	/* 1 milisecond */
		}
	}
	return NULL;
}

static FILE *
open_status_file(pid_t pid)
{
	PROC_PID_FILE(fn, "/proc/%d/status", pid);
	/* Don't complain if we fail.  This would typically happen
	   when the process is about to terminate, and these files are
	   not available anymore.  This function is called from the
	   event loop, and we don't want to clutter the output just
	   because the process terminates.  */
	return fopen(fn, "r");
}

static char *
find_line_starting(FILE * file, const char * prefix, size_t len)
{
	char * line = NULL;
	size_t line_len = 0;
	while (!feof(file)) {
		if (getline(&line, &line_len, file) < 0)
			return NULL;
		if (strncmp(line, prefix, len) == 0)
			return line;
	}
	return NULL;
}

static void
each_line_starting(FILE * file, const char *prefix,
		   enum pcb_status (*cb)(const char * line, const char * prefix,
					 void * data),
		   void * data)
{
	size_t len = strlen(prefix);
	char * line;
	while ((line = find_line_starting(file, prefix, len)) != NULL) {
		enum pcb_status st = (*cb)(line, prefix, data);
		free (line);
		if (st == pcb_stop)
			return;
	}
}

static enum pcb_status
process_leader_cb(const char * line, const char * prefix, void * data)
{
	pid_t * pidp = data;
	*pidp = atoi(line + strlen(prefix));
	return pcb_stop;
}

pid_t
process_leader(pid_t pid)
{
	pid_t tgid = 0;
	FILE * file = open_status_file(pid);
	if (file != NULL) {
		each_line_starting(file, "Tgid:\t", &process_leader_cb, &tgid);
		fclose(file);
	}

	return tgid;
}

static enum pcb_status
process_stopped_cb(const char * line, const char * prefix, void * data)
{
	char c = line[strlen(prefix)];
	// t:tracing stop, T:job control stop
	*(int *)data = (c == 't' || c == 'T');
	return pcb_stop;
}

int
process_stopped(pid_t pid)
{
	int is_stopped = -1;
	FILE * file = open_status_file(pid);
	if (file != NULL) {
		each_line_starting(file, "State:\t", &process_stopped_cb,
				   &is_stopped);
		fclose(file);
	}
	return is_stopped;
}

static enum pcb_status
process_status_cb(const char * line, const char * prefix, void * data)
{
	const char * status = line + strlen(prefix);
	const char c = *status;

#define RETURN(C) do {					\
		*(enum process_status *)data = C;	\
		return pcb_stop;			\
	} while (0)

	switch (c) {
	case 'Z': RETURN(ps_zombie);
	case 't': RETURN(ps_tracing_stop);
	case 'T':
		/* This can be either "T (stopped)" or, for older
		 * kernels, "T (tracing stop)".  */
		if (!strcmp(status, "T (stopped)\n"))
			RETURN(ps_stop);
		else if (!strcmp(status, "T (tracing stop)\n"))
			RETURN(ps_tracing_stop);
		else {
			fprintf(stderr, "Unknown process status: %s",
				status);
			RETURN(ps_stop); /* Some sort of stop
					  * anyway.  */
		}
	case 'D':
	case 'S': RETURN(ps_sleeping);
	}

	RETURN(ps_other);
#undef RETURN
}

enum process_status
process_status(pid_t pid)
{
	enum process_status ret = ps_invalid;
	FILE * file = open_status_file(pid);
	if (file != NULL) {
		each_line_starting(file, "State:\t", &process_status_cb, &ret);
		fclose(file);
		if (ret == ps_invalid)
			error(0, errno, "process_status %d", pid);
	} else
		/* If the file is not present, the process presumably
		 * exited already.  */
		ret = ps_zombie;

	return ret;
}

static int
all_digits(const char *str)
{
	while (isdigit(*str))
		str++;
	return !*str;
}

int
process_tasks(pid_t pid, pid_t **ret_tasks, size_t *ret_n)
{
	PROC_PID_FILE(fn, "/proc/%d/task", pid);
	DIR * d = opendir(fn);
	if (d == NULL)
		return -1;

	/* XXX This is racy.  We need to stop the tasks that we
	   discover this way and re-scan the directory to eventually
	   reach a full set of tasks.  */
	pid_t *tasks = NULL;
	size_t n = 0;
	size_t alloc = 0;

	while (1) {
		struct dirent entry;
		struct dirent *result;
		if (readdir_r(d, &entry, &result) != 0) {
		fail:
			closedir(d);
			free(tasks);
			return -1;
		}
		if (result == NULL)
			break;
		if (result->d_type == DT_DIR && all_digits(result->d_name)) {
			pid_t npid = atoi(result->d_name);
			if (n >= alloc) {
				alloc = alloc > 0 ? (2 * alloc) : 8;
				pid_t *ntasks = realloc(tasks,
							sizeof(*tasks) * alloc);
				if (ntasks == NULL)
					goto fail;
				tasks = ntasks;
			}
			if (n >= alloc)
				abort();
			tasks[n++] = npid;
		}
	}

	closedir(d);

	*ret_tasks = tasks;
	*ret_n = n;
	return 0;
}

int
task_kill (pid_t pid, int sig)
{
	// Taken from GDB
        int ret;

        errno = 0;
        ret = syscall (__NR_tkill, pid, sig);
	return ret;
}

static int
fetch_dyn64(struct process *proc, void **addr, Elf64_Dyn *ret)
{
	if (umovebytes(proc, *addr, ret, sizeof(*ret)) != sizeof(*ret))
		return -1;
	*addr += sizeof(*ret);
	return 0;
}

static int
fetch_dyn32(struct process *proc, void **addr, Elf64_Dyn *ret)
{
	Elf32_Dyn dyn;
	if (umovebytes(proc, *addr, &dyn, sizeof(dyn)) != sizeof(dyn))
		return -1;

	*addr += sizeof(dyn);
	ret->d_tag = dyn.d_tag;
	ret->d_un.d_val = dyn.d_un.d_val;

	return 0;
}

/* On native 64-bit system, we need to be careful when handling cross
 * tracing.  This select appropriate pointer depending on host and
 * target architectures.  */
static void *
select_32_64(struct process *proc, void *p32, void *p64)
{
	if (sizeof(long) == 4 || proc->mask_32bit)
		return p32;
	else
		return p64;
}

static int (*
dyn_fetcher(struct process *proc))(struct process *,
				   void **, Elf64_Dyn *)
{
	return select_32_64(proc, fetch_dyn32, fetch_dyn64);
}

static int
find_dynamic_entry_addr(struct process *proc, int d_tag, void **addr)
{
	debug(1, "find_dynamic_entry()");

	if (addr == NULL || proc->dyn_addr == NULL
	    || d_tag < 0 || d_tag > DT_NUM)
		return -1;

	int i = 0;
	void *dyn_addr = proc->dyn_addr;
	while (1) {
		Elf64_Dyn entry;
		if (dyn_fetcher(proc)(proc, &dyn_addr, &entry) < 0
		    || entry.d_tag == DT_NULL
		    || i++ > ELF_MAX_SEGMENTS) {
			debug(2, "Couldn't find address for dtag!");
			return -1;
		}

		if (entry.d_tag == d_tag) {
			*addr = (void *)(uintptr_t)entry.d_un.d_val;
			debug(2, "found address: %p, dtag %d", *addr, d_tag);
			return 0;
		}
	}
}

struct cb_data {
	const char *lib_name;
	GElf_Addr bias;
	struct process *proc;
	int is_base;
};
enum {
	LT_RD_BASE = 0x1000,
};

/* Our own type for representing 32-bit linkmap.  We can't rely on the
 * definition in link.h, because that's only accurate for our host
 * architecture, not for target architecture (where the traced process
 * runs). */
#define LT_LINK_MAP(BITS)			\
	{					\
		Elf##BITS##_Addr l_addr;	\
		Elf##BITS##_Addr l_name;	\
		Elf##BITS##_Addr l_ld;		\
		Elf##BITS##_Addr l_next;	\
		Elf##BITS##_Addr l_prev;	\
	}

struct lt_link_map_32 LT_LINK_MAP(32);
struct lt_link_map_64 LT_LINK_MAP(64);

#undef LT_LINK_MAP

static int
fetch_lm64(struct process *proc, Elf64_Addr addr, struct lt_link_map_64 *ret)
{
	if (umovebytes(proc, (void *)(uintptr_t)addr, ret,
		       sizeof(*ret)) != sizeof(*ret))
		return -1;
	return 0;
}

static int
fetch_lm32(struct process *proc, Elf64_Addr addr, struct lt_link_map_64 *ret)
{
	struct lt_link_map_32 lm;
	if (umovebytes(proc, (void *)(uintptr_t)addr, &lm,
		       sizeof(lm)) != sizeof(lm))
		return -1;

	ret->l_addr = lm.l_addr;
	ret->l_name = lm.l_name;
	ret->l_ld = lm.l_ld;
	ret->l_next = lm.l_next;
	ret->l_prev = lm.l_prev;

	return 0;
}

static int (*
lm_fetcher(struct process *proc))(struct process *,
				  Elf64_Addr, struct lt_link_map_64 *)
{
	return select_32_64(proc, fetch_lm32, fetch_lm64);
}

/* The same as above holds for struct r_debug.  */
#define LT_R_DEBUG(BITS)			\
	{					\
		int r_version;			\
		Elf##BITS##_Addr r_map;		\
		Elf##BITS##_Addr r_brk;		\
		int r_state;			\
		Elf##BITS##_Addr r_ldbase;	\
	}

struct lt_r_debug_32 LT_R_DEBUG(32);
struct lt_r_debug_64 LT_R_DEBUG(64);

#undef LT_R_DEBUG

static int
fetch_rd64(struct process *proc, Elf64_Addr addr, struct lt_r_debug_64 *ret)
{
	if (umovebytes(proc, (void *)(uintptr_t)addr, ret,
		       sizeof(*ret)) != sizeof(*ret))
		return -1;
	return 0;
}

static int
fetch_rd32(struct process *proc, Elf64_Addr addr, struct lt_r_debug_64 *ret)
{
	struct lt_r_debug_32 rd;
	if (umovebytes(proc, (void *)(uintptr_t)addr, &rd,
		       sizeof(rd)) != sizeof(rd))
		return -1;

	ret->r_version = rd.r_version;
	ret->r_map = rd.r_map;
	ret->r_brk = rd.r_brk;
	ret->r_state = rd.r_state;
	ret->r_ldbase = rd.r_ldbase;

	return 0;
}

static int (*
rdebug_fetcher(struct process *proc))(struct process *,
				      Elf64_Addr, struct lt_r_debug_64 *)
{
	return select_32_64(proc, fetch_rd32, fetch_rd64);
}

static void
crawl_linkmap(struct process *proc, struct lt_r_debug_64 *dbg,
	      void (*callback)(void *), struct cb_data *data)
{
	debug (1, "crawl_linkmap()");

	if (!dbg || !dbg->r_map) {
		debug(2, "Debug structure or it's linkmap are NULL!");
		return;
	}

	Elf64_Addr addr = dbg->r_map;

	while (addr != 0) {
		struct lt_link_map_64 rlm;
		if (lm_fetcher(proc)(proc, addr, &rlm) < 0) {
			debug(2, "Unable to read link map\n");
			return;
		}

		addr = rlm.l_next;
		if (rlm.l_name == 0) {
			debug(2, "Invalid library name referenced in dynamic linker map\n");
			return;
		}

		char lib_name[BUFSIZ];
		umovebytes(proc, (void *)(uintptr_t)rlm.l_name,
			   lib_name, sizeof(lib_name));

		if (lib_name[0] == '\0') {
			debug(2, "Library name is an empty string");
			continue;
		}

		if (callback) {
			debug(2, "Dispatching callback for: %s, bias 0x%x\n",
			      lib_name, rlm.l_addr);

			/* Note that l_addr is a _bias_, not absolute
			 * load address.  It also happens to actually
			 * be base address (as the comment in link.h
			 * implies) unless the file is prelinked.  */
			data->bias = rlm.l_addr;
			data->lib_name = lib_name;
			callback(data);
		}
	}
	return;
}

static int
load_debug_struct(struct process *proc, struct lt_r_debug_64 *ret)
{
	debug(1, "load_debug_struct");

	if (rdebug_fetcher(proc)(proc, (uintptr_t)proc->debug, ret) < 0) {
		debug(2, "This process does not have a debug structure!\n");
		return -1;
	}

	return 0;
}

static GElf_Addr
bias(struct process *proc, GElf_Addr base, GElf_Addr bias)
{
	/* We need to respect the bitness of operands so that result
	 * wraps around the right way.  */
	if (sizeof(long) == 4 || proc->mask_32bit)
		return (Elf32_Addr)base + (Elf32_Addr)bias;
	else
		return base + bias;
}

static int
deref_opd(struct process *proc, GElf_Addr raddr, GElf_Addr *retp)
{
#ifdef __powerpc64__
	if (!proc->mask_32bit) {
		if (umovebytes(proc, (void *)raddr, retp, sizeof(*retp))
		    != sizeof(*retp))
			return -1;

		return 0;
	}
#endif

	*retp = raddr;
	return 0;
}

static int
add_library_breakpoint(struct cb_data *lm_add, const char *name, GElf_Sym *sym)
{
	debug(2, "found symbol %s @ %lx, adding it.", name, sym->st_value);

	GElf_Addr addr = bias(lm_add->proc, sym->st_value, lm_add->bias);
	GElf_Addr orig = addr;
	if (deref_opd(lm_add->proc, addr, &addr) < 0) {
		fprintf(stderr, "Couldn't resolve OPD for %s\n", name);
		return -1;
	}

	/* When reading OPD of the initial bunch of libraries, we have
	 * already let the dynamic linker run (the _r_debug probe is
	 * installed after we hit _start).  But later on, the probe
	 * hits as soon as the linker calls _dl_debug_state, and
	 * that's before the OPD is resolved.  For the former case, we
	 * have to avoid applying bias.  */
	if (addr != orig && !lm_add->is_base)
		addr = bias(lm_add->proc, addr, lm_add->bias);

	int indirect = GELF_ST_TYPE(sym->st_info) == STT_GNU_IFUNC;
	char *xn = (char *)name;
	if (indirect && asprintf(&xn, "%s.IFUNC", name) < 0)
		return -1;

	add_library_symbol(addr, xn, &library_symbols,
			   LS_TOPLT_NONE, 0);
	if (indirect)
		free(xn); /* add_library_symbol makes a copy itself.  */
	library_symbols->is_indirect = indirect;

	insert_breakpoint(lm_add->proc,
			  sym2addr(lm_add->proc, library_symbols),
			  library_symbols, 1);
	return 0;
}

/* Put a breakpoint to all symbols of this library.  */
static void
populate_dl(struct cb_data *lm_add, struct ltelf *lte)
{
	const char *libname
		= lte->soname != NULL ? lte->soname : lm_add->lib_name;
	size_t i;
	for (i = 0; i < lte->dynsym_count; ++i) {
		GElf_Sym sym;
		if (gelf_getsym(lte->dynsym, i, &sym) == NULL)
			continue;
		int bind = GELF_ST_BIND(sym.st_info);
		int type = GELF_ST_TYPE(sym.st_info);
		if ((bind != STB_GLOBAL && type != STB_LOCAL)
		    || (type != STT_FUNC && type != STT_GNU_IFUNC)
		    || sym.st_shndx == SHN_UNDEF)
			continue;

		const char *name = lte->dynstr + sym.st_name;
		char *rename;
		if (asprintf(&rename, "%s@%s", name, libname) < 0)
			rename = (char *)name;
		add_library_breakpoint(lm_add, rename, &sym);
		if (rename != name)
			/* add_library_symbol makes a copy itself.  */
			free(rename);
	}
}

/* Populate a library with breakpoints requested with -x.  */
static void
populate_x(struct cb_data *lm_add, struct ltelf *lte)
{
	GElf_Sym sym;
	struct opt_x_t *xptr;
	for (xptr = opt_x; xptr; xptr = xptr->next)
		if (!xptr->found
		    && in_load_libraries(xptr->name, lte, 1, &sym)
		    && add_library_breakpoint(lm_add, xptr->name, &sym) == 0)
			xptr->found = 1;
}

static void
linkmap_add_cb(void *data) { //const char *lib_name, ElfW(Addr) addr) {
	int i = 0;
	struct cb_data *lm_add = data;

	debug(1, "linkmap_add_cb");

	/*
		XXX
		iterate through library[i]'s to see if this lib is in the list.
		if not, add it
	 */
	for(;i < library_num;i++) {
		if (strcmp(library[i], lm_add->lib_name) == 0) {
			/* found it, so its not new */
			return;
		}
	}

	/* new library linked! */
	debug(2, "New libdl loaded library found: %s\n", lm_add->lib_name);

	if (library_num < MAX_LIBRARY) {
		struct ltelf lte;
		library[library_num++] = strdup(lm_add->lib_name);
		memset(&lte, 0, sizeof(struct ltelf));

		do_init_elf(&lte, library[library_num-1]);

		/* IS_BASE is true for libraries that the binary
		 * depends on, opened during the first dynamic linker
		 * pass.  It is false for libraries opened with
		 * dlopen.
		 *
		 * We populate all libraries with static breakpoints.
		 * When in --dl mode, for !base libraries we _instead_
		 * place breakpoints on _all_ static symbols, as if -x
		 * was requested for each of them.  */
		if (lm_add->is_base || !opt_dl)
			populate_x(lm_add, &lte);
		else
			populate_dl(lm_add, &lte);
		do_close_elf(&lte);
	}
}

void
arch_check_dbg(struct process *proc)
{
	debug(1, "arch_check_dbg");

	struct lt_r_debug_64 dbg;
	if (load_debug_struct(proc, &dbg) < 0) {
		debug(2, "Unable to load debug structure!");
		return;
	}

	long prev_state = proc->debug_state;
	int is_base = 0;
	if (prev_state & LT_RD_BASE) {
		prev_state &= ~LT_RD_BASE;
		is_base = 1;
	}

	if (dbg.r_state == RT_CONSISTENT) {
		debug(2, "Linkmap is now consistent");
		if (prev_state == RT_ADD) {
			struct cb_data data = {
				.proc = proc,
				.is_base = is_base,
			};
			debug(2, "Adding DSO to linkmap");
			crawl_linkmap(proc, &dbg, linkmap_add_cb, &data);
		} else if (prev_state == RT_DELETE) {
			debug(2, "Removing DSO from linkmap");
		} else {
			debug(2, "Unexpected debug state!");
		}
	}

	proc->debug_state = dbg.r_state;
}

static void
hook_libdl_cb(void *data) {
	struct cb_data *hook_data = data;

	debug(1, "add_library_cb");

	if (!data) {
		debug(2, "No callback data");
		return;
	}

	const char *lib_name = hook_data->lib_name;

	if (hook_data->proc->interp != NULL
	    && strcmp(hook_data->proc->interp, lib_name) != 0) {
		linkmap_add_cb(hook_data);
		return;
	}

	if (library_num < MAX_LIBRARY) {
		library[library_num++] = strdup(lib_name);
	}
	else {
		fprintf (stderr, "MAX LIBS REACHED\n");
		exit(EXIT_FAILURE);
	}
}

int
linkmap_init(struct process *proc)
{
	debug(1, "linkmap_init()");

	if (find_dynamic_entry_addr(proc, DT_DEBUG, &proc->debug) == -1) {
		debug(2, "Couldn't find debug structure!");
		return -1;
	}

	struct lt_r_debug_64 rdbg;
	if (load_debug_struct(proc, &rdbg) < 0) {
		debug(2, "No debug structure or no memory to allocate one!");
		return -1;
	}

	GElf_Addr faddr;
	if (deref_opd(proc, rdbg.r_brk, &faddr) < 0) {
		fprintf(stderr, "Can't load r_brk address from OPD\n");
		return -1;
	}

	add_library_symbol(faddr, "", &library_symbols, LS_TOPLT_NONE, 0);
	insert_breakpoint(proc, sym2addr(proc, library_symbols), library_symbols, 1);

	struct cb_data data = {
		.proc = proc,
		.is_base = 1,
	};
	crawl_linkmap(proc, &rdbg, hook_libdl_cb, &data);
	proc->debug_state = RT_ADD | LT_RD_BASE;

	return 0;
}
