#if HAVE_CONFIG_H
#include "config.h"
#endif

#define	_GNU_SOURCE	1
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>
#include <unistd.h>
#include <sys/ptrace.h>

#include "ltrace.h"
#include "options.h"
#include "debug.h"
#include "elf.h"

#define FORK_MASK 0x00010000

/* If the system headers did not provide the constants, hard-code the normal
 *    values.  */
#ifndef PTRACE_OLDSETOPTIONS
#define PTRACE_OLDSETOPTIONS    21
#endif
#ifndef PTRACE_SETOPTIONS
#define PTRACE_SETOPTIONS       0x4200
#endif
#ifndef PTRACE_GETEVENTMSG
#define PTRACE_GETEVENTMSG      0x4201
#endif

/* options set using PTRACE_SETOPTIONS */
#ifndef PTRACE_O_TRACESYSGOOD
#define PTRACE_O_TRACESYSGOOD   0x00000001
#endif
#ifndef PTRACE_O_TRACEFORK
#define PTRACE_O_TRACEFORK      0x00000002
#endif
#ifndef PTRACE_O_TRACEVFORK
#define PTRACE_O_TRACEVFORK     0x00000004
#endif
#ifndef PTRACE_O_TRACECLONE
#define PTRACE_O_TRACECLONE     0x00000008
#endif
#ifndef PTRACE_O_TRACEEXEC
#define PTRACE_O_TRACEEXEC      0x00000010
#endif
#ifndef PTRACE_O_TRACEVFORKDONE
#define PTRACE_O_TRACEVFORKDONE 0x00000020
#endif
#ifndef PTRACE_O_TRACEEXIT
#define PTRACE_O_TRACEEXIT      0x00000040
#endif

/* Wait extended result codes for the above trace options.  */
#ifndef PTRACE_EVENT_FORK
#define PTRACE_EVENT_FORK       1
#endif
#ifndef PTRACE_EVENT_VFORK
#define PTRACE_EVENT_VFORK      2
#endif
#ifndef PTRACE_EVENT_CLONE
#define PTRACE_EVENT_CLONE      3
#endif
#ifndef PTRACE_EVENT_EXEC
#define PTRACE_EVENT_EXEC       4
#endif
#ifndef PTRACE_EVENT_VFORK_DONE
#define PTRACE_EVENT_VFORK_DONE 5
#endif
#ifndef PTRACE_EVENT_EXIT
#define PTRACE_EVENT_EXIT       6
#endif

static struct event event;

/* This should also update `current_process' */

struct process *pid2proc(int pid);
void verify(int sig_status);

/* A queue of events that we missed while enabling the
 * breakpoint in one of tasks.  */
static struct event * delayed_events = NULL;
static struct event * end_delayed_events = NULL;

static enum pcb_status
first(struct process * proc, void * data)
{
	return pcb_stop;
}

void
enque_event(struct event * event)
{
	struct event * ne = malloc(sizeof(*ne));
	if (ne == NULL) {
		perror("event will be missed: malloc");
		return;
	}

	*ne = *event;
	ne->next = NULL;
	if (end_delayed_events == NULL) {
		assert(delayed_events == NULL);
		end_delayed_events = delayed_events = ne;
	}
	else {
		assert(delayed_events != NULL);
		end_delayed_events = end_delayed_events->next = ne;
	}
}

struct event *
each_qd_event(enum ecb_status (*pred)(struct event *, void *), void * data)
{
	struct event * prev = delayed_events;
	struct event * event;
	for (event = prev; event != NULL; ) {
		switch ((*pred)(event, data)) {
		case ecb_cont:
			prev = event;
			event = event->next;
			continue;

		case ecb_deque:
			if (end_delayed_events == event)
				end_delayed_events = prev;
			if (delayed_events == event)
				delayed_events = event->next;
			else
				prev->next = event->next;
			if (delayed_events == NULL)
				end_delayed_events = NULL;
			/* fall-through */

		case ecb_yield:
			return event;
		}
	}

	return NULL;
}

static enum ecb_status
event_process_not_reenabling(struct event * event, void * data)
{
	if (event->proc == NULL
	    || event->proc->leader == NULL
	    || event->proc->leader->event_handler == NULL)
		return ecb_deque;
	else
		return ecb_cont;
}

static struct event *
next_qd_event(void)
{
	return each_qd_event(&event_process_not_reenabling, NULL);
}

static int in_waitpid = 0;

int
should_call_ltrace_exiting(void)
{
	/* If IN_WAITPID is set, we are in or around waitpid call in
	 * wait_for_something.  That means that we can touch process
	 * handlers and such.  It also means that unless we do it
	 * right away, we will wait until one of the processes goes
	 * through an event, and that can be forever, because waitpid
	 * doesn't get interrupted for SIGINT.  */
	return in_waitpid;
}

struct event *wait_for_something(void)
{
	pid_t pid;
	int status;
	int tmp;
	int stop_signal;

	struct event * ev;
	if ((ev = next_qd_event()) != NULL) {
		event = *ev;
		free(ev);
		return &event;
	}

	if (!each_process(NULL, &first, NULL)) {
		debug(1, "No more children");
		exit(0);
	}

	in_waitpid = 1;
	pid = waitpid(-1, &status, __WALL);
	in_waitpid = 0;

	if (pid == -1) {
		if (errno == ECHILD) {
			debug(1, "No more children");
			exit(0);
		} else if (errno == EINTR) {
			debug(1, "wait received EINTR ?");
			event.thing = LT_EV_NONE;
			return &event;
		}
		perror("wait");
		exit(1);
	}

	debug(3, "signal from pid %u, status %#x", pid, status);
	event.proc = pid2proc(pid);
	if (!event.proc || event.proc->state == STATE_BEING_CREATED) {
		/* On ia64, we are seeing a waitpid event even though
		 * the process is still reported to be running.  Wait
		 * for the tracing stop to propagate.  But don't get
		 * stuck here forever.  */
		int i = 0;
		for (; i < 100 && process_status(pid) != ps_tracing_stop; ++i) {
			debug(2, "waiting for %d to stop", pid);
			usleep(10000);
		}
		event.thing = LT_EV_NEW;
		event.e_un.newpid = pid;
		return &event;
	}
	get_arch_dep(event.proc);

	struct process *leader = event.proc->leader;
	if (leader == event.proc) {
		if (!event.proc->libdl_hooked) {
			/* debug struct may not have been written yet.. */
			if (linkmap_init(event.proc) == 0) {
				event.proc->libdl_hooked = 1;
			}
		}
	}

	/* The process should be stopped after the waitpid call.  But
	 * when the whole thread group is terminated, we see
	 * individual tasks spontaneously transitioning from 't' to
	 * 'R' and 'Z'.  Calls to ptrace fail and /proc/pid/status may
	 * not even be available anymore, so we can't check in
	 * advance.  So we just drop the error checking around ptrace
	 * calls.  We check for termination ex post when it fails,
	 * suppress the event, and let the event loop collect the
	 * termination in the next iteration.  */
#define CHECK_PROCESS_TERMINATED					\
	do {								\
		int errno_save = errno;					\
		switch (process_stopped(pid))				\
		case 0:							\
		case -1: {						\
			event.thing = LT_EV_NONE;			\
			continue_process(event.proc->pid);		\
			return &event;					\
		}							\
		errno = errno_save;					\
	} while (0)

	event.proc->instruction_pointer = (void *)(uintptr_t)-1;

	/* Check for task termination now, before we have a need to
	 * call CHECK_PROCESS_TERMINATED later.  That would suppress
	 * the event that we are processing.  */
	if (WIFEXITED(status)) {
		event.thing = LT_EV_EXIT;
		event.e_un.ret_val = WEXITSTATUS(status);
		return &event;
	}
	if (WIFSIGNALED(status)) {
		event.thing = LT_EV_EXIT_SIGNAL;
		event.e_un.signum = WTERMSIG(status);
		return &event;
	}

	event.proc->instruction_pointer = get_instruction_pointer(event.proc);
	if (event.proc->instruction_pointer == (void *)(uintptr_t)-1) {
		CHECK_PROCESS_TERMINATED;
		if (errno != 0)
			perror("get_instruction_pointer");
 	}
	switch (syscall_p(event.proc, status, &tmp)) {
	case 1:
		event.thing = LT_EV_SYSCALL;
		event.e_un.sysnum = tmp;
		return &event;
	case 2:
		event.thing = LT_EV_SYSRET;
		event.e_un.sysnum = tmp;
		return &event;
	case -1:
		CHECK_PROCESS_TERMINATED;
		if (errno != 0)
			perror("syscall_p");
	}
	if (WIFSTOPPED(status)) {
		int what = status >> 16;
		if (what == PTRACE_EVENT_VFORK
		    || what == PTRACE_EVENT_FORK
		    || what == PTRACE_EVENT_CLONE) {
			unsigned long data;
			event.thing = what == PTRACE_EVENT_VFORK
				? LT_EV_VFORK : LT_EV_CLONE;
			ptrace(PTRACE_GETEVENTMSG, pid, NULL, &data);
			event.e_un.newpid = data;
			return &event;
		}
	}
	/* TODO: check for EVENT_CLONE */
	if (WIFSTOPPED(status) && (status>>16 == PTRACE_EVENT_EXEC)) {
		event.thing = LT_EV_EXEC;
		return &event;
	}
	if (!WIFSTOPPED(status)) {
		event.thing = LT_EV_UNKNOWN;
		return &event;
	}

	stop_signal = WSTOPSIG(status);
	event.thing = LT_EV_NONE;

	/* On some targets, breakpoints are signalled not using
	   SIGTRAP, but also with SIGILL, SIGSEGV or SIGEMT.  Check
	   for these. */
	if (stop_signal == SIGSEGV
	    || stop_signal == SIGILL
#ifdef SIGEMT
	    || stop_signal == SIGEMT
#endif
	    ) {
		// If we didn't need to know IP so far, get it now.
		void * addr = opt_i
		  ? event.proc->instruction_pointer
		  : (event.proc->instruction_pointer = get_instruction_pointer (event.proc));

		if (address2bpstruct(leader, addr - DECR_PC_AFTER_BREAK))
			stop_signal = SIGTRAP;
	}

	if (event.thing == LT_EV_NONE
	    && (stop_signal != (SIGTRAP | event.proc->tracesysgood))
	    && (stop_signal != SIGTRAP)) {
		event.thing = LT_EV_SIGNAL;
		event.e_un.signum = stop_signal;
		return &event;
	}

	event.thing = LT_EV_BREAKPOINT;
	event.e_un.brk_addr =
	    event.proc->instruction_pointer - DECR_PC_AFTER_BREAK;
	return &event;
}

static enum pcb_status
find_proc(struct process * proc, void * data)
{
	pid_t pid = (pid_t)(uintptr_t)data;
	return proc->pid == pid ? pcb_stop : pcb_cont;
}

struct process *
pid2proc(pid_t pid)
{
	return each_process(NULL, &find_proc, (void *)(uintptr_t)pid);
}
